from typing import Dict
from unittest import mock

import pkg_resources as pkg
import pytest

# import before skidward.models to avoid circular imports
from skidward import web
from skidward import JobStatus
from skidward.backend import RedisBackend
from skidward.models import db, Job, Task, Worker
from worker_detector.namespace_module_manager import NamespaceModuleManager


@pytest.fixture(autouse=True)
def enable_testing(monkeypatch):
    monkeypatch.setenv("TESTING", "True")
    monkeypatch.setenv("SQLALCHEMY_DATABASE_URI", "sqlite:///:memory:")


@pytest.fixture()
def init_database():
    db.create_all()
    worker_one = Worker(name="real_name")
    db.session.add(worker_one)
    db.session.commit()

    task_one = Task(
        name="task_1",
        worker=worker_one,
        context={"setting_1": "value_1"},
        cron_string="*/2 * * * *",
    )
    db.session.add(task_one)
    db.session.commit()
    yield db
    db.session.remove()
    db.drop_all()


@pytest.fixture
def real_namespace() -> str:
    return "skidward.real.name.space"


class MockEntryPoint(pkg.EntryPoint):
    def __init__(self, name, module_name):
        self.name = name
        self.module_name = module_name
        self.attrs = []
        self.extras = []
        self.dist = []

    def load(self, *args, **kwargs) -> pkg.EntryPoint:
        return MockEntryPoint(self.name, self.module_name)

    def start(self, context):
        return True


class MockNamespaceModuleManager(NamespaceModuleManager):
    def _get_module_entry_points_on_namespace(self) -> Dict[str, pkg.EntryPoint]:
        if self.namespace == "skidward.real.name.space":
            return {
                "real_name": MockEntryPoint("real-name", "real_name"),
                "other_name": MockEntryPoint("other-name", "other_name"),
            }

        return super()._get_module_entry_points_on_namespace()


@pytest.fixture
def mock_namespace_module_manager() -> MockNamespaceModuleManager:
    return MockNamespaceModuleManager("skidward.real.name.space")


@pytest.fixture()
def backend():
    dummy_backend = RedisBackend()
    yield dummy_backend
    dummy_backend.dummy_backend.erase()



import datetime
from unittest import mock

from skidward.models import Job, JobStatus, Task, Worker
from skidward.task_runner import TaskRunner


class TestTaskRunner:
    def test_job_status_is_updated(self, init_database):
        task = Task.query.all()[0]
        job = Job(state=JobStatus.READY, ran_at=datetime.datetime.utcnow(), task=task)
        init_database.session.add(job)
        init_database.session.commit()
        TaskRunner().update_job_status(job, JobStatus.RUNNING)

        assert job.state == JobStatus.RUNNING

    def test_new_job_is_created(self, init_database):
        worker = Worker.query.all()[0]
        context = {"LOGGING_CHANNEL": "ch:channel"}
        task = Task(name="task_id", worker=worker, context=context, cron_string="")
        init_database.session.add(task)
        init_database.session.commit()
        TaskRunner().create_new_job(task)

        assert init_database.session.query(Job).one().task.id == task.id

    @mock.patch("worker_detector.create_namespace_module_manager")
    def test_full_process_calling_run(
        self, mock_create, real_namespace, init_database, mock_namespace_module_manager
    ):
        worker = Worker.query.all()[0]
        context = {"LOGGING_CHANNEL": "ch:channel"}
        task = Task(name="task_id", worker=worker, context=context, cron_string="")
        init_database.session.add(task)
        init_database.session.commit()

        TaskRunner().start(task)
        job = init_database.session.query(Job).one()

        mock_create.return_value.load_module.assert_called_once_with("real_name")
        mock_create.return_value.load_module.return_value.start.assert_called_once_with(
            {"LOGGING_CHANNEL": "ch:channel"}
        )
        assert job.id == 1
        assert job.task_id == 2
        assert job.state == JobStatus.SUCCESS
        assert isinstance(job.ran_at, datetime.datetime)



import datetime
import logging

from redis_log_handler.RedisLogHandler import RedisLogHandler
from skidward.models import db, Job, JobStatus
import worker_detector


class TaskRunner:
    def update_job_status(self, job, status):
        job.query.update({"state": status})
        db.session.commit()

    def create_new_job(self, task):
        job = Job(
            state=JobStatus.RUNNING, ran_at=datetime.datetime.utcnow(), task_id=task.id
        )
        db.session.add(job)
        db.session.commit()
        logging.info("Job id:{} created".format(job.id))
        return job

    def start(self, task):
        worker_name = task.worker.name
        context = task.context

        # could get channel from context .ie {'LOGGING_CHANNEL' : "ch:channel"}
        log_handler = RedisLogHandler(context["LOGGING_CHANNEL"])
        logging.basicConfig(handlers=(log_handler,), level=logging.INFO)

        logging.info("Logging {}".format(worker_name))

        job = self.create_new_job(task)
        worker_module = worker_detector.load_worker_on_namespace(worker_name)

        logging.info("{} is running".format(worker_name))

        try:
            worker_module.start(context)
            status = JobStatus.SUCCESS
        except:
            status = JobStatus.FAIL

        logging.info("Status : {}".format(status))

        self.update_job_status(job, status=status)

        log_handler.close()
